﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Reflection;
using System.Web;
using System.Web.UI;
using System.Web.UI.WebControls;
using DymWebForm.Common.Linq;
using DymWebForm.Common.Web;
using DymWebForm.Common.BasicDataType;

namespace DymWebForm
{
    public class DymWebFormBuilder
    {
        private readonly FormField[] _allField;
        private readonly FormField _pkField;
        private readonly FormFieldFactory _formFieldFactory;
        private readonly Dictionary<Type, FormFieldFactory> _dicFormFieldFactory;

        public Type TableType { private set; get; }


        public DymWebFormBuilder(Type type)
        {
            TableType = type;
            _formFieldFactory = new FormFieldFactory(type);
            _allField = _formFieldFactory.GetFormFields();
            _pkField = _allField.FirstOrDefault(p => p.IsPK);

            if (_pkField == null)
                throw new ArgumentException("Must define PK");

            _dicFormFieldFactory = new Dictionary<Type, FormFieldFactory>();
            _dicFormFieldFactory[type] = _formFieldFactory;
        }

        private void ProcTdFK(Func<Type, object[], object[]> fkQuery, List<FieldFK> lstFKs)
        {
            lstFKs
                .GroupBy(p => p.Field.FKParentType)
                .ToDictionary(p => p.Key, p => p.ToArray())
                .ForEach(p =>
                {

                    var ff = _dicFormFieldFactory.ContainsKey(p.Key)
                        ? _dicFormFieldFactory[p.Key]
                        : (_dicFormFieldFactory[p.Key] = new FormFieldFactory(p.Key));

                    var arrFk = p.Value
                        .Select(q => q.FKVal)
                        .Distinct()
                        .ToArray();

                    var dicFKobj = fkQuery(p.Key, arrFk)
                        .ToDictionary(
                            ff.GetPKValue
                            , q => q
                        );

                    p.Value.ForEach(q =>
                    {
                        q.Td.Controls.Add(new Literal { Text = q.FKVal.ToDisplayString() });
                        q.Td.Controls.Add(new Literal { Text = ": " });

                        object parVal = null;
                        var found = false;
                        if (dicFKobj.ContainsKey(q.FKVal))
                        {
                            found = true;
                            var par = dicFKobj[q.FKVal];
                            parVal = ff.GetFKValue(par);
                        }

                        if (found)
                        {
                            q.Td.AddControl(new Label { Text = parVal.ToDisplayString() ?? "NULL", ForeColor = Color.Blue })
                                         .Font.Italic = (parVal == null);
                        }
                        else
                        {
                            q.Td.Controls.Add(new Label { Text = "无对应", ForeColor = Color.Red });
                        }
                    });
                });
        }

        private Table RenderDetailsInternal(WebControl container, object entity, Func<Type, object[], object[]> fkQuery, string[] skipFields = null)
        {
            skipFields = skipFields ?? new string[] { };

            var arrFieldsToDisplay = _allField
                    .Where(p => skipFields.Contains(p.FieldName) == false)
                    .ToArray();

            //init
            var tbl = new Table();
            container.Controls.Add(tbl);

            tbl.Attributes.Add("border", "1");
            tbl.Style[HtmlTextWriterStyle.BorderCollapse] = "Collapse";
            tbl.CellSpacing = 1;
            tbl.CellPadding = 2;

            var lstFKs = new List<FieldFK>();

            //generate fields of form
            foreach (var field in arrFieldsToDisplay)
            {
                var tr = new TableRow();
                tbl.Rows.Add(tr);

                var objVal = field.GetValue(entity);

                //field name
                tr.Cells.Add(new TableCell { Text = field.DisplayName + ":" });

                //field control
                var td = new TableCell();
                tr.Cells.Add(td);
                td.VisualText(objVal.ToDisplayString());

                //fk
                if (field.IsFK && objVal != null)
                {
                    lstFKs.Add(new FieldFK { Field = field, Td = td, FKVal = objVal });
                }
            }

            if (fkQuery != null)
            {
                ProcTdFK(fkQuery, lstFKs);
            }
            return tbl;
        }


        ///  <summary>
        /// 
        ///  </summary>
        ///  <param name="container">容器, 要求已经添加到Page</param>
        ///  <param name="dataSource">数据源</param>
        ///  <param name="arrSkipFields">跳过的字段(如密码)</param>
        ///  <param name="fkQuery">外键查询回调</param>
        ///  <param name="urlEdit">{0}=ID 默认 ?op=edit&amp;arg={0}</param>
        ///  <param name="urlDelete">{0}=ID 默认 ?op=delete&amp;arg={0}</param>
        /// <param name="pageIndex"></param>
        /// <param name="pageSize"></param>
        public void RenderGrid(Control container, IEnumerable dataSource, string[] arrSkipFields
            , Func<Type, object[], object[]> fkQuery = null
            , string urlEdit = null
            , string urlDelete = null
            , int pageIndex = 0, int pageSize = 0)
        {
            //TODO: should allow optional columns to show
            
            //init vars
            urlEdit = urlEdit ?? "?op=edit&arg={0}";
            urlDelete = urlDelete ?? "?op=delete&arg={0}";

            //init
            var tbl = new Table();
            container.Controls.Add(tbl);

            tbl.Attributes.Add("border", "1");
            tbl.Style[HtmlTextWriterStyle.BorderCollapse] = "Collapse";
            tbl.CellSpacing = 1;
            tbl.CellPadding = 2;

            arrSkipFields = arrSkipFields ?? new string[] { };

            var arrFieldsToDisplay = _allField
                    .Where(p => arrSkipFields.Contains(p.FieldName) == false)
                    .ToArray();

            //header
            var th = new TableRow();

            th.Cells.Add(new TableHeaderCell { Text = "#" });

            th.Cells.AddRange(
                arrFieldsToDisplay
                    .SelectArray(p => (TableCell)new TableHeaderCell { Text = p.DisplayName })

            );

            th.Cells.Add(new TableHeaderCell { Text = "操作" });

            tbl.Rows.Add(th);

            //coll fk td for batch process
            var lstFKs = new List<FieldFK>();

            var rowNumber = pageIndex * pageSize;

            //rows
            foreach (var item in dataSource)
            {
                //TODO: optimize, use pure html tag(not server control)
                var tr = new TableRow();
                tbl.Rows.Add(tr);

                tr.Cells.Add(new TableCell { Text = (++rowNumber).ToString() });

                //entity data
                var item1 = item;
                tr.Cells.AddRange(
                    arrFieldsToDisplay
                        .SelectArray(p =>
                        {
                            var objVal = p.GetValue(item1);

                            var strVal = objVal.ToDisplayString();

                            var td = new TableCell();

                            td.VisualText(strVal);

                            if (p.IsFK && objVal != null)
                            {
                                lstFKs.Add(new FieldFK { Field = p, Td = td, FKVal = objVal });
                            }

                            return td;
                        })
                );

                //opt.
                var tdOpt = new TableCell();
                tr.Cells.Add(tdOpt);

                var key = _pkField.GetValue(item);

                tdOpt.Controls.Add(new HyperLink { Text = "修改", NavigateUrl = urlEdit.Format(key) });
                tdOpt.Controls.Add(new Literal { Text = " " });
                tdOpt.Controls.Add(new HyperLink { Text = "删除", NavigateUrl = urlDelete.Format(key) });
            }//end foreach

            if (fkQuery != null)
            {
                ProcTdFK(fkQuery, lstFKs);
            }
        }

        public void RenderInputForm(WebControl container, Action actOnSubmit, bool isAdd = true)
        {
            //init
            var tbl = new Table();
            container.Controls.Add(tbl);

            tbl.Attributes.Add("border", "1");
            tbl.Style[HtmlTextWriterStyle.BorderCollapse] = "Collapse";
            tbl.CellSpacing = 1;
            tbl.CellPadding = 2;

            //generate fields of form
            foreach (var field in _allField)
            {
                if (isAdd && field.IsPK)
                    continue;

                var validation = field.Validation;

                var tr = new TableRow();
                tbl.Rows.Add(tr);

                //field name
                var tdName = tr.AddCell(new TableCell { Text = field.DisplayName + ":" });
                if (field.AllowNull == false)
                {
                    tdName.Controls.Add(new Literal { Text = "[" });
                    tdName.Controls.Add(new Label { Text = "*", ForeColor = Color.Red });
                    tdName.Controls.Add(new Literal { Text = "]" });
                    tdName.Controls.Add(new Literal { Text = field.DisplayName + ":" });
                }

                //field control
                var tdCtl = new TableCell();
                tr.Cells.Add(tdCtl);

                var ctl = field.CreateControl();

                if (field.IsPK)
                {
                    var box = ctl as TextBox;
                    if (box != null)
                    {
                        var tc = box;
                        tc.ReadOnly = true;
                    }
                }

                if (field.HasDefaultValueGenerator)
                    DymWebFormControlMappnig.GetInst(field.DataType, field.ValidationType).ControlSetValue(ctl, field.GenerateDefaultValue());

                tdCtl.Controls.Add(ctl);
                ctl.ID = field.ControlID;

                //hint
                if (!isAdd && validation is DynWebFormPasswordValidation)
                {
                    tdCtl.Controls.Add(new Literal { Text = "<br>*留空保持不变" });
                    tdName.VerticalAlign = VerticalAlign.Top;
                }

                //field validate
                var tdVal = new TableCell();
                tr.Cells.Add(tdVal);

                if (field.IsPK == false
                    && field.AllowNull == false
                    && ctl is CheckBox == false
                    && (
                        (isAdd)
                        || (false == validation is DynWebFormPasswordValidation)
                    )
                )
                {
                    var rfv = new RequiredFieldValidator();
                    tdVal.Controls.Add(rfv);

                    rfv.ValidationGroup = "fc" + TableType.Name;
                    rfv.ControlToValidate = ctl.ID;

                    rfv.Display = ValidatorDisplay.Dynamic;
                    rfv.ForeColor = Color.Red;
                    rfv.Text = "*";
                    rfv.ErrorMessage = "必填字段:<b>" + field.DisplayName + "</b>";
                    rfv.ID = "rfv" + ctl.ID;
                }

                if (validation != null)
                {
                    BaseValidator vc = validation.CreateValidator(container);

                    tdVal.Controls.Add(vc);

                    vc.ValidationGroup = "fc" + TableType.Name;
                    vc.ControlToValidate = ctl.ID;

                    vc.Display = ValidatorDisplay.Dynamic;
                    vc.ForeColor = Color.Red;
                    vc.Text = "▲";
                    vc.ErrorMessage = "无效输入:<b>" + field.DisplayName + "</b>";
                    vc.ID = "cv" + ctl.ID;
                }
            }

            //the submit button
            var trSubmit = new TableRow();
            tbl.Rows.Add(trSubmit);

            var tdSubmit = new TableCell();
            trSubmit.Cells.Add(tdSubmit);
            tdSubmit.ColumnSpan = 3;
            tdSubmit.HorizontalAlign = HorizontalAlign.Right;

            tdSubmit.AddControl(new Button
            {
                Text = "提交",
                ValidationGroup = "fc" + TableType.Name
            }).Click += (ss, ee) =>
            {
                if (((Page)HttpContext.Current.Handler).IsValid)
                    actOnSubmit();
            };

            //the summary
            var trSummarny = new TableRow();
            tbl.Rows.Add(trSummarny);

            var tdSummary = new TableCell();
            trSummarny.Cells.Add(tdSummary);

            tdSummary.ColumnSpan = 3;

            tdSummary.Controls.Add(new ValidationSummary { ValidationGroup = "fc" + TableType.Name, ForeColor = Color.Red });
        }

        public void BindForm(WebControl container, object entity, bool isAdd = false)
        {
            _allField.ForEach(p =>
            {
                var val = p.GetValue(entity);
                var octl = container.FindControl(p.ControlID);
                DymWebFormControlMappnig.GetInst(p.DataType, p.ValidationType).ControlSetValue(octl, val);
            });
        }

        public void BindEntity(WebControl container, object entity)
        {
            if (container == null)
                throw new ArgumentNullException("container");

            _allField
                .Where(p => p.IsPK == false)
                .ForEach(p =>
                {
                    var octl = container.FindControl(p.ControlID);
                    var val = DymWebFormControlMappnig.GetInst(p.DataType, p.ValidationType).ControlGetValue(octl);

                    if (p.Validation is DynWebFormPasswordValidation
                            && (val == null || val as string == "")
                        )
                        return;

                    p.SetValue(entity, val);
                });
        }

        public void RenderDetails(WebControl container, object entity, Func<Type, object[], object[]> fkQuery = null, string[] arrSkipFields = null)
        {
            RenderDetailsInternal(container, entity, fkQuery, arrSkipFields);
        }

        public void RenderDelete(WebControl container, object entity, Action actOnDeleteSubmit, Func<Type, object[], object[]> fkQuery = null, string[] skipFields = null)
        {
            var tbl = RenderDetailsInternal(container, entity, fkQuery, skipFields);

            //the submit button
            var trSubmit = new TableRow();
            tbl.Rows.Add(trSubmit);

            var tdSubmit = new TableCell();
            trSubmit.Cells.Add(tdSubmit);
            tdSubmit.ColumnSpan = 3;
            tdSubmit.HorizontalAlign = HorizontalAlign.Right;

            tdSubmit.AddControl(new Button
            {
                Text = "删除",
            }).Click += (ss, ee) =>
            {
                if (((Page)HttpContext.Current.Handler).IsValid)
                    actOnDeleteSubmit();
            };
        }

        public Dictionary<string, string> GetFieldNamesForGrid()
        {
            return _allField
                .Where(p => false == p.Validation is DynWebFormPasswordValidation)
                .ToDictionary(p => p.FieldName, p => p.DisplayName);
        }

        public PropertyInfo GetFieldInfoForGrid(string name)
        {
            return _allField
                .Where(p => p.FieldName == name && false == p.Validation is DynWebFormPasswordValidation)
                .Select(p => p.PropertyInfo)
                .FirstOrDefault();
        }
    }
}
